"""Saku Gateway base module.
"""
#
# Copyright (c) 2005,2006 shinGETsu Project.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Id: gateway.py 1204 2006-06-04 11:45:48Z fuktommy $
#

import cgi
import os.path
import re
import urllib
import sys
import time
if hasattr(sys, "winver"):
    from os import listdir
else:
    from dircache import listdir

import basecgi
import config
import spam
from cache import *
from node import *

__version__ = "$Revision: 1204 $"


class Message(dict):

    """Multi-language message for gateway."""

    def __init__(self, file):
        dict.__init__(self)
        try:
            f = open(file)
            del_eos = re.compile(r"[\r\n]*")
            iscomment = re.compile(r"^#$").search
            for line in f:
                line = del_eos.sub("", line)
                if iscomment(line):
                    pass
                else:
                    buf = line.split("<>")
                    if len(buf) == 2:
                        buf[1] = urllib.unquote(buf[1])
                        self[buf[0]] = buf[1]
            f.close()
        except IOError:
            sys.stderr.write(file + ": IOError\n")

# End of Message


def search_message(accept_language):
    """Search message file.

    Example of accept_language: "ja,en-us;q=0.7,en;q=0.3"
    """
    q = {}
    lang = []
    if accept_language != "":
        for i in accept_language.split(","):
            found = re.search(r"(\S+)\s*;\s*q=(\S+)", i)
            if found:
                try:
                    lang = found.group(1)
                    q[lang] = float(found.group(2))
                except ValueError:
                    pass
            else:
                q[i] = 1

        lang = q.keys()
        lang.sort(lambda a,b: cmp(q[b], q[a]))
    lang.append(config.language)
    for i in lang:
        file = config.file_dir + "/" + "message-" + i + ".txt"
        if os.path.isfile(file):
            return Message(file)
    return None

# End of search_message


class CGI(basecgi.CGI):

    root = config.root_path
    sep = config.query_separator
    appli = config.application
    gateway_cgi = config.gateway
    thread_cgi = config.thread_cgi
    admin_cgi = config.admin_cgi
    message = None
    filter = None
    str_filter = ''

    def __init__(self,
                 stdin=sys.stdin,
                 stdout=sys.stdout,
                 stderr=sys.stderr,
                 environ=os.environ):
        basecgi.CGI.__init__(self,
                             stdin=stdin,
                             stdout=stdout,
                             stderr=stderr,
                             environ=environ)
        if "HTTP_ACCEPT_LANGUAGE" in self.environ:
            al = self.environ["HTTP_ACCEPT_LANGUAGE"]
        else:
            al = ""
        self.message = search_message(al)
        addr = self.environ.get("REMOTE_ADDR", "")
        self.isadmin = re.search(config.admin, addr)
        self.isfriend = re.search(config.friend, addr)
        self.isvisitor = re.search(config.visitor, addr)

    def path_info(self):
        """Parse PATH_INFO.

        If PATH_INFO is not defined, use QUERY_STRING.
        x.cgi?foo&bar=y -> path="foo".
        """
        m = re.search(r"([^&;=]*)(&|$)", self.environ.get("QUERY_STRING", ""))
        if self.environ.get("PATH_INFO", "") != "":
            path = self.environ["PATH_INFO"]
            if path.startswith("/"):
                path = path[1:]
        elif m is not None:
            path = m.group(1)
        else:
            path = ""
        path = self.escape(self.str_decode(path))
        return path

    def str_encode(self, query):
        """Encode for URI.

        ~ -> %7E
        """
        return urllib.quote(query)

    def str_decode(self, query):
        """Decode URI.

        %7E -> ~
        """
        return urllib.unquote(query)

    def file_encode(self, type, query):
        """Encode for filename.

        ~ -> 7E
        """
        buf = [type, "_"]
        for i in query:
            buf.append("%02X" % ord(i))
        return "".join(buf)

    def file_decode(self, query):
        """Decode filename.

        7E -> ~
        """
        q = query.split("_")
        if q < 1:
            return None
        else:
            query = q[1]
        buf = []
        for i in range(0, len(query), 2):
            try:
                buf.append("%c" % int(query[i:i+2], 16))
            except (ValueError, IndexError):
                self.stderr.write(query + ": ValueError/IndexError\n")
                return None
        return "".join(buf)

    def escape(self, msg):
        msg = msg.replace("&", "&amp;")
        msg = re.sub(r"&amp;(#\d+|#[Xx][0-9A-Fa-f]+|[A-Za-z0-9]+);",
                     r"&\1;",
                     msg)
        msg = msg.replace("<", "&lt;")
        msg = msg.replace(">", "&gt;")
        msg = msg.replace("\r", "")
        msg = msg.replace("\n", "<br>")
        return msg

    def gateway_link(self, cginame, command):
        return '<a href="%s%s%s" title="%s">%s</a>' % \
               (cginame, self.sep, command,
                self.message.get('desc_'+command, ''), self.message[command])

    def extension(self, suffix):
        filename = []
        for i in listdir(config.abs_docroot):
            if i.endswith('.%s' % suffix) and \
               (not (i.startswith('.') or i.startswith('_'))):
                filename.append(i)
        filename.sort()
        return filename

    def menubar(self, id='', rss=''):
        if id:
            id = ' id="%s"' % id
        if rss == "":
            rss = self.gateway_cgi + self.sep + "rss"
        message = self.message
        buf = []
        self.stdout.write(
            '<p class="menu"%s>\n' % id +
            '  <a href="' + self.gateway_cgi + '">' +
            message["top"] + "</a> |\n")
        for i in ("changes",):
            buf.append('  ' + self.gateway_link(self.gateway_cgi, i))
        if self.isfriend or self.isadmin:
            for i in ("recent", "new"):
                buf.append('  ' + self.gateway_link(self.gateway_cgi, i))
        buf.append('  <a href="' + rss + '">' + message["rss"] + "</a>")
        self.stdout.write(" |\n".join(buf) + "\n</p>\n")

    def header(self, title='', rss='', cookie=None, deny_robot=False):
        """Print CGI and HTTP header."""
        if rss == "":
            rss = self.gateway_cgi + self.sep + "rss"
        message = self.message
        self.stdout.write("Content-Type: text/html; charset=UTF-8\r\n")
        if cookie:
            self.stdout.write("%s\r\n" % cookie)
        self.stdout.write(
            "\r\n" +
            '<?xml version="1.0" encoding="UTF-8"?>\n' +
            '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"\n'+
            '  "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">\n' +
            '<html xmlns="http://www.w3.org/1999/xhtml"' +
            ' lang="%s" xml:lang="%s">\n' %
                (message["lang"], message["lang"]) +
            '<head>\n' +
            '  <meta http-equiv="content-type"' +
            ' content="text/html; charset=UTF-8" />\n' +
            '  <meta http-equiv="content-script-type"' +
            ' content="text/javascript" />\n' +
            "  <title>" + title + "</title>\n")
        if deny_robot:
            self.stdout.write('  <meta name="robots" content="NOINDEX" />\n')
        self.stdout.write(
            '  <link rev="made" href="http://shingetsu.info/" />\n'+
            '  <link rel="contents" href="/" />\n' +
            '  <link rel="alternate" type="application/rss+xml" title="RSS"' +
            ' href="' + rss + '" />\n' +
            '  <script type="text/javascript">\n' +
            '    var uiLang = "%s";\n' % message['lang'] +
            '    var initFunc = [];\n' +
            '    function doInitFunc() {\n' +
            '      for (var i=0; i!=initFunc.length; i++) {\n' +
            '        initFunc[i]();\n' +
            '      }\n' +
            '    }\n' +
            '  </script>\n')
        for e in self.extension('js'):
            self.stdout.write(
                '  <script type="text/javascript" src="%s%s"></script>\n'
                    % (config.root_path, e))
        for e in self.extension('css'):
            self.stdout.write(
            '  <link rel="stylesheet" type="text/css" href="%s%s" />\n'
                    % (config.root_path, e))
        self.stdout.write(
            '</head>\n' +
            '<body onload="doInitFunc()">\n')
        self.menubar("top", rss=rss)
        if self.isadmin:
            self.stdout.write(
                '<h1><a href="%s/search%s%s" title="%s">' %
                (self.admin_cgi, self.sep,
                 self.str_encode(title), message["search"]) +
                '%s</a></h1>\n' % title)
        else:
            self.stdout.write('<h1>%s</h1>\n' % title)

    def footer(self):
        self.stdout.write(
            '<address>Powered by ' +
            '<a href="http://shingetsu.info/">shinGETsu</a>.</address>\n')
        self.stdout.write("</body>\n</html>\n")

    def localtime(self, stamp=0):
        """Return YYYY-mm-dd HH:MM."""
        return time.strftime('%Y-%m-%d %H:%M', time.localtime(int(stamp)))

    def rfc822_time(self, stamp=0):
        """Return date and time in RFC822 format."""
        return time.strftime("%a, %d %b %Y %H:%M:%S GMT",
                             time.gmtime(int(stamp)))

    def res_anchor(self, id):
        return ('<a href="#r%s" class="innerlink">') % id

    def html_format(self, plain):
        buf = plain.replace("<br>", "\n")
        buf = self.escape(buf)
        buf = re.sub(r"(&gt;&gt;)([0-9a-f]{8})",
                     self.res_anchor(r"\2") + r"\g<0></a>",
                     buf)
        buf = re.sub(r"https?://[^\x00-\x20\"'()<>\[\]\x7F-\xFF]{2,}",
                     r'<a href="\g<0>">\g<0></a>',
                     buf)
        buf = re.sub(r'\[\[<a.*?>(.*?)\]\]</a>', r'[[\1]]', buf)
        buf = re.sub(r'  ', '&nbsp;&nbsp;', buf)
        buf = re.sub(r'<br> ', '<br>&nbsp;', buf)
        buf = re.sub(r'^ ', '&nbsp;', buf)
        buf = re.sub(r'<br>', '<br />\n    ',buf)

        tmp = ""
        while buf:
            m = re.search(r"\[\[([^<>]+?)\]\]", buf)
            if m is not None:
                tmp += buf[:m.start()]
                tmp += self.bracket_link(m.group(1))
                buf = buf[m.end():]
            else:
                tmp += buf
                buf = ""
        return tmp

    def bracket_link(self, link):
        """Encode bracket string to link.

        See WikiWikiWeb.
        """
        m = re.search(r"^/(thread)/([^/]+)/([0-9a-f]{8})$", link)
        if m is not None:
            uri = self.thread_cgi + self.sep + \
                  self.str_encode(m.group(2)) + \
                  "#r" + m.group(3)
            return '<a href="' + uri + '">[[' + link + ']]</a>'

        m = re.search(r"^/(thread)/([^/]+)$", link)
        if m is not None:
            uri = self.appli[m.group(1)] + self.sep + \
                  self.str_encode(m.group(2))
            return '<a href="' + uri + '">[[' + link + ']]</a>'

        m = re.search(r"^([^/]+)/([0-9a-f]{8})$", link)
        if m is not None:
            uri = self.appli[self.appli_type] + self.sep + \
                  self.str_encode(m.group(1)) + \
                  "#r" + m.group(2)
            return '<a href="' + uri + '">[[' + link + ']]</a>'

        m = re.search(r"^([^/]+)$", link)
        if m is not None:
            uri = self.appli[self.appli_type] + self.sep + \
                  self.str_encode(m.group(1))
            return '<a href="' + uri + '">[[' + link + ']]</a>'

        return "[[" + link + "]]"

    def remove_file_form(self, cache):
        filesize = str(cache.size*10/1024/1024/10.0)
        if self.isadmin:
            self.stdout.write(
                '<form method="post" action="' + self.admin_cgi + '"><p>\n' +
                '  <input type="submit" value="' + self.message["del_file"] + 
                '" tabindex="9" accesskey="d" />\n' +
                '  <input type="hidden" name="cmd" value="fdel" />\n' +
                '  <input type="hidden" name="file" value="' +
                cache.datfile + '" />\n' +
                "  %s%s\n" % (filesize, self.message["mb"]) +
                "</p></form>\n")
        else:
            self.stdout.write("<p>%s%s</p>\n" % (filesize, self.message["mb"]))

    def print302(self, next):
        """Print CGI header (302 moved temporarily)."""
        self.header("Loading...")
        self.stdout.write(
            '<p>Click and jump to <a href="' + next + '">' + next + '</a>' +
            '</p>\n' +
            '<script type="text/javascript">' +
            'window.location.href = "' + next + '";</script>\n')
        self.footer()

    def print403(self):
        """Print CGI header (403 forbidden)."""
        self.header(self.message["403"], deny_robot=True)
        self.stdout.write("<p>" + self.message["403_body"] + "</p>")
        self.footer()

    def print404(self, cache=None):
        """Print CGI header (404 not found)."""
        self.header(self.message["404"], deny_robot=True)
        self.stdout.write("<p>" + self.message["404_body"] + "</p>")
        if cache is not None:
            self.remove_file_form(cache)
        self.footer()

    def lock(self):
        if self.isadmin:
            lockfile = config.admin_search
        else:
            lockfile = config.search_lock
        if not os.path.isfile(lockfile):
            f = file(lockfile, "wb")
            f.close()
            return True
        elif os.path.getmtime(lockfile) + config.search_timeout < time.time():
            f = file(lockfile, "wb")
            f.close()
            return True
        else:
            return False

    def unlock(self):
        if self.isadmin:
            lockfile = config.admin_search
        else:
            lockfile = config.search_lock
        try:
            os.remove(lockfile)
        except (OSError, IOError), err:
            self.stderr.write("%s: OSError/IOError: %s\n" % (lockfile, err))
            return False

    def get_cache(self, cache):
        """Search cache from network."""
        result = cache.search()
        self.unlock()
        return result

    def print_new_element_form(self, parent=None):
        if not (self.isadmin or self.isfriend):
            return
        else:
            datfile = ""
            cginame = self.gateway_cgi

        self.stdout.write(
            '<form method="post" action="' + cginame + '"><p>\n' +
            '  <input type="hidden" name="cmd" value="new" />\n' +
            '  <input type="hidden" name="file" value="' + datfile + '" />\n' +
            '  <input type="submit" value="' + self.message["new"] + '"' +
            ' name="submit" tabindex="1" accesskey="w" />\n')

        #for type in config.types:
        #    self.stdout.write(
        #        '  <input type="radio" value="' + type + '" name="type"' +
        #        ' tabindex="2" accesskey="t" />' +
        #        self.message[type] + "\n")
        self.stdout.write(
            '  <input type="hidden" value="thread" name="type" />\n')

        #if parent is not None:
        #    self.stdout.write(
        #        '  <input type="checkbox" name="dopost" value="dopost"' +
        #        ' checked="checked" tabindex="2" accesskey="s" />' +
        #        self.message["send"] + "\n")

        self.stdout.write(
            '  ' + self.message["title"] + ": " +
            '<input name="link" size="40" value=""' +
            ' tabindex="3" accesskey="t" />\n' +
            "</p></form>\n")

    def error_time(self):
        from random import gauss
        return int(gauss(time.time(), config.time_error))

    def do_post(self, path, form):
        """Post article."""
        import base64
        try:
            attach = form["attach"]
        except KeyError:
            attach = None
            str_attach = ""

        if (attach is not None) and attach.file:
            if (len(attach.value) > config.file_limit*1024*1024):
                self.header(self.message["big_file"], deny_robot=True)
                self.footer()
                return None
            str_attach = base64.encodestring(attach.value).replace("\n", "")
        guess_suffix = "txt"
        if (attach is not None) and attach.filename != "":
            found = re.search(r"\.([^.]+)$", attach.filename)
            if found:
                guess_suffix = found.group(1).lower()

        suffix = form.getfirst("suffix", "")
        if (suffix == "") or (suffix == "AUTO"):
            suffix = guess_suffix
        elif suffix.startswith("."):
            suffix = suffix[1:].lower()
        else:
            suffix = suffix.lower()
        suffix = re.sub(r"[^0-9A-Za-z]", "", suffix)

        if form.getfirst("error", "") != "":
            stamp = self.error_time()
        else:
            stamp = int(time.time())

        body = {}
        for key in ("base_stamp", "base_id", "name", "mail", "body"):
            value = form.getfirst(key, "")
            if value != "":
                body[key] = self.escape(value)
        if str_attach != "":
            body["attach"] = str_attach
            body["suffix"] = re.sub(r"[\r\n]", "", suffix)

        if not body:
            self.header(self.message["null_article"], deny_robot=True)
            self.footer()
            return None

        cache = Cache(form.getfirst("file"))
        rec = Record(datfile=cache.datfile)
        passwd = form.getfirst("passwd", "")
        id = rec.build(stamp, body, passwd=passwd)

        if spam.check(rec.recstr):
            self.header(self.message['spam'], deny_robot=True)
            self.footer()
            return None

        if cache.exists():
            cache.add_data([rec])
        else:
            self.print404()
            return None

        if form.getfirst("dopost", "") != "":
            updatelist = UpdateList()
            updatelist.append(rec)
            updatelist.sync()
            recentlist = RecentList()
            recentlist.append(rec)
            recentlist.sync()
            nodelist = NodeList()
            nodelist.tell_update(cache, stamp=stamp, id=id)

        return id[:8]

    def check_get_cache(self):
        agent = self.environ.get("HTTP_USER_AGENT", "")
        if not (self.isfriend or self.isadmin):
            return False
        elif re.search(config.robot, agent):
            return False
        elif self.lock():
            return True
        else:
            return False

    def check_visitor(self):
        return self.isadmin or self.isfriend or self.isvisitor

    def make_list_item(self, cache, remove=True):
        x = self.file_decode(cache.datfile)
        y = self.str_encode(x)
        try:
            u = unicode(x, 'utf-8')
        except UnicodeDecodeError:
            return None
        if self.filter and (not self.filter.search(u)):
            return None
        buf = "  <li>\n"
        if self.isadmin and remove:
            buf += ('    <input type="checkbox" name="file" value="%s"' +
                    ' tabindex="1" accesskey="s" />\n') % cache.datfile
        buf += (
            '    <span class="stamp" id="s%d">%s</span>\n' %
                (cache.stamp, self.localtime(cache.stamp)) +

            '    <a href="%s%s%s">%s</a>\n' % 
                (self.appli[cache.type], self.sep, y, x) +
            '    (%s/%.1f%s)\n' %
                (len(cache), cache.size/1024/1024.0, self.message["mb"]) +
            '  </li>\n')
        return buf

    def print_index_list(self, cachelist, target="", footer=True):
        """Print index list."""
        if target:
            self.stdout.write(
                '<form method="get" action="%s%s%s"><p>' %
                (self.gateway_cgi, self.sep, target) +
                '  <input type="submit" value="%s" '%self.message['filter'] + 
                'accesskey="g" tabindex="10" />\n' +
                '  ' + self.message['regexp'] + ': ' +
                '<input name="filter" size="40" value="%s" '%self.str_filter +
                'accesskey="f" tabindex="9" />' +
                '</p></form>\n')
        if self.isadmin:
            self.stdout.write(
                '<form method="post" action="%s">\n' % self.admin_cgi +
                '<p><input type="hidden" name="cmd" value="fdel" /></p>\n')
        for type in config.types:
            self.stdout.write("<ul>\n")
            for cache in cachelist:
                if cache.type == type:
                    try:
                        buf = self.make_list_item(cache)
                        if buf:
                            self.stdout.write(buf)
                    except KeyError:
                        pass
            self.stdout.write("</ul>\n")
        if self.isadmin:
            self.stdout.write(
                '<p><input type="submit" value="' +
                self.message["del_file"] + '" tabindex="2" accesskey="d" />' +
                '</p>\n' +
                "</form>\n")
        if target:
            cols = "file,stamp,date,path,uri,type,title,records,size"
            self.stdout.write('<p><a href="%s%scsv/%s/%s">CSV</a></p>' %
                              (self.gateway_cgi, self.sep, target, cols))
        if footer:
            self.print_new_element_form();
            self.footer()
# End of CGI
